# -*- coding: utf-8 -*-

import threading
import time
import zmq

from . import misc


class ZMQObservationSenderThread(threading.Thread):
    """Sends observations via zmq sockets."""
    def __init__(self, thread_id, thread_name, output_buffer, done_buffer, config, exit_event):
        threading.Thread.__init__(self)
        self.thread_id = thread_id
        self.name = thread_name
        self.output_buffer = output_buffer
        self.done_buffer = done_buffer
        self.exit_event = exit_event
        self._socket = config['ObservationSender']['socket']
        self._update_interval = config['ObservationSender']['update_interval']
        self._init_zmq_sockets()

    def _init_zmq_sockets(self):
        """Initializes and binds all zmq sockets"""
        self.context = zmq.Context()

        # Socket to send output
        self.output_socket = self.context.socket(zmq.PUB)
        self.output_socket.bind('tcp://*:{socket}'.format(socket=self._socket))

    def _send_output(self):
        """Sends the output buffer"""
        ts = time.time()
        for item in self.output_buffer[0]:
            item.update({'ts': ts})
        self.output_socket.send_json(self.output_buffer[0])

    def run(self):
        while not self.exit_event.is_set():
            t_start = time.time()
            self._send_output()
            misc.sleep_remaining(t_start, self._update_interval, 'ObservationSenderThread: sleep time negative')
        print '[INFO] ObservationSenderThread shutting down.'

    def done(self):
        return self.done_buffer[0]



# -*- coding: utf-8 -*-

class AleWrapper(WrapperBase):
    """Wrapper for the ale library"""
    def __init__(self, rom_path):
        raise NotImplementedError('Emulator currently not supported.')
        EmuWrapper.__init__(self)

        self._rom_path = os.path.expanduser(rom_path)
        self._ale = ale_python_interface.ALEInterface()

        # Set USE_SDL to true to display the screen. ALE must be compiled
        # with SDL enabled for this to work.
        USE_SDL = True
        if USE_SDL:
            self._ale.setBool('sound', False)
            self._ale.setBool('display_screen', True)

    def seed(self, seed):
        self._ale.setInt(b'random_seed', seed)

    def load_env(self, env, *args, **kwargs):
        rom = glob.glob(os.path.join(self._rom_path, env) + '*')
        if rom:
            self._ale.loadROM(rom[0])
        else:
            raise ValueError('Environment not found.')

    def get_output_dimensions(self, as_tuple=False):
        if as_tuple:
            return (self._ale.getScreenDims()[1], self._ale.getScreenDims()[0])
        else:
            return self._ale.getScreenDims()[0] * self._ale.getScreenDims()[1]

    def reset(self):
        self._ale.reset_game()

    def execute_action(self, action):
        self._reward = self._ale.act(action[0])
        self._done_buffer[0] = self._ale.game_over()
        return self._reward

    def update_output_buffer(self):
        assert(self._output_buffer is not None)
        self._output_buffer[0] = list(np.array(self._ale.getScreen().flatten(), dtype=np.float))

    def get_command_buffer(self):
        if self._command_buffer is None:
            self._command_buffer = [[0]]
        return self._command_buffer

    def get_output_buffer(self):
        if self._output_buffer is None:
            self._output_buffer = [list(np.zeros(self.get_output_dimensions(), dtype=np.float))]
        return self._output_buffer




# -*- coding: utf-8 -*-

from .zmq_reward_sender_thread import ZMQRewardSenderThread
from .zmq_observation_sender_thread import ZMQObservationSenderThread
from .zmq_command_receiver_thread import ZMQCommandReceiverThread
from .env_runner_thread import EnvRunnerThread
from .gym_wrapper import GymWrapper

__version__ = '0.0.1'



# -*- coding: utf-8 -*-

import collections
import json
import os
import sys
import time
import warnings


def read_default_config():
    """Reads the default configuration file"""
    with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'DefaultConfig.json'), 'r') as f:
        return json.load(f)


def recursively_update_dict(d, u):
    """Updates dict d with dict u recursively."""
    for k, v in u.iteritems():
        if isinstance(v, collections.Mapping):
            r = recursively_update_dict(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def sleep_remaining(t_start, t_total, msg=''):
    """Sleeps the remaining time from now to t_start + t_total."""
    t_end = time.time()
    if t_total > (t_end - t_start):
        time.sleep(t_total - (t_end - t_start))
    else:
        with warnings.catch_warnings():
            warnings.simplefilter('always')  # always show desyncing warnings
            warnings.warn(msg, RuntimeWarning)


class SignalHandler(object):
    """
    Handles SIGINT by setting exit event and waiting for threads to
    finish.
    See https://christopherdavis.me/blog/threading-basics.html
    """

    def __init__(self, exit_event, threads):
        self.exit_event = exit_event
        self.threads = threads

    def __call__(self, signum, frame):
        # Set exit event for all threads
        self.exit_event.set()

        # Wait for all threads to finish
        for thread in self.threads:
            thread.join()

        # And exit the program
        sys.exit(0)



# -*- coding: utf-8 -*-

from . import messages


class WrapperBase(object):
    """Base class for all wrappers"""

    def __init__(self):
        self._command_buffer = None
        self._output_buffer = None
        self._reward_buffer = None
        self._done_buffer = None

        self._gym = None
        self._output = None

        self._reward = None
        self._initial_reward = None
        self._min_reward = None
        self._max_reward = None

        self._inter_trial_observation = None

    def done(self):
        return self._done_buffer[0]

    def seed(self, seed):
        """Sets random seed of environment."""
        raise NotImplementedError()

    def load_env(self, env, *args, **kwargs):
        """Loads a specified environment."""
        raise NotImplementedError()

    def reset(self):
        """Resets environment to initial state."""
        raise NotImplementedError()

    def execute_action(self):
        """Executes a single action in an environment."""
        raise NotImplementedError()

    def update_output_buffer(self, data):
        """Updates output buffer with current observation."""
        raise NotImplementedError()

    def update_reward_buffer(self):
        """Updates reward buffer with current reward."""
        assert(self._reward_buffer is not None)
        self._reward_buffer[0] = messages.to_message(self._min_reward, self._max_reward, self._reward)

    def get_command_buffer(self):
        """Initializes command buffer."""
        raise NotImplementedError()

    def get_done_buffer(self):
        """Initializes buffer to signal end of episode."""
        if self._done_buffer is None:
            self._done_buffer = [False]
        return self._done_buffer

    def get_output_buffer(self):
        """Initializes output buffer."""
        raise NotImplementedError()

    def get_reward_buffer(self):
        """Initializes reward buffer."""
        if self._reward_buffer is None:
            self._reward_buffer = [messages.to_message(self._min_reward, self._max_reward, self._initial_reward)]
        return self._reward_buffer

    def clear_output_buffer(self):
        """Replaces observations in output buffer with user defined values."""
        raise NotImplementedError()

    def clear_reward_buffer(self):
        """Replaces reward in reward buffer with user defined values."""
        self._reward_buffer[0] = messages.to_message(self._min_reward, self._max_reward, self._initial_reward)

    def report(self):
        """Writes a custom report."""
        raise NotImplementedError()



# -*- coding: utf-8 -*-

import numpy as np


def to_message(low, high, value):
    assert(np.shape(low) == np.shape(high) == np.shape(value))
    try:
        return [{'min': low[i], 'max': high[i], 'value': value[i]} for i in xrange(len(low))]
    except TypeError:
        return [{'min': low, 'max': high, 'value': value}]



# -*- coding: utf-8 -*-

import threading
import time
import zmq

from . import misc


class ZMQRewardSenderThread(threading.Thread):
    """Sends rewards via zmq sockets."""

    def __init__(self, thread_id, thread_name, reward_buffer, done_buffer, config, exit_event):
        threading.Thread.__init__(self)
        self.thread_id = thread_id
        self.name = thread_name
        self._buffer = reward_buffer
        self.done_buffer = done_buffer
        self.exit_event = exit_event
        self._socket = config['RewardSender']['socket']
        self._update_inteval = config['RewardSender']['update_interval']
        self._init_zmq_sockets()

    def _init_zmq_sockets(self):
        """Initializes and binds all zmq sockets"""
        self.context = zmq.Context()

        # Socket to send output
        self.socket = self.context.socket(zmq.PUB)
        self.socket.bind('tcp://*:{socket}'.format(socket=self._socket))

    def _send(self):
        """Sends the output buffer"""
        ts = time.time()
        for item in self._buffer[0]:
            item.update({'ts': ts})
        self.socket.send_json(self._buffer[0])

    def run(self):
        while not self.exit_event.is_set():
            t_start = time.time()
            self._send()
            misc.sleep_remaining(t_start, self._update_inteval, 'RewardSenderThread: sleep time negative')
        print '[INFO] RewardSenderThread shutting down.'

    def done(self):
        return self.done_buffer[0]



# -*- coding: utf-8 -*-

import threading
import time
import warnings
import zmq


class ZMQCommandReceiverThread(threading.Thread):
    """Receives command via zmq socket."""

    def __init__(self, thread_id, thread_name, command_buffer, done_buffer, config, exit_event):
        threading.Thread.__init__(self)

        self.thread_id = thread_id
        self.name = thread_name
        self.command_buffer = command_buffer
        self.done_buffer = done_buffer
        self.exit_event = exit_event
        self._socket = config['CommandReceiver']['socket']
        self._time_stamp_tolerance = config['CommandReceiver']['time_stamp_tolerance']
        self._init_zmq_sockets()

    def _init_zmq_sockets(self):
        """Initializes and binds all zmq sockets"""
        self.context = zmq.Context()

        # Socket to receive commands
        self.command_socket = self.context.socket(zmq.SUB)
        self.command_socket.RCVTIMEO = 1000  # set timeout to allow for checking exit event
        self.command_socket.connect('tcp://localhost:{socket}'.format(socket=self._socket))
        self.command_socket.setsockopt(zmq.SUBSCRIBE, "")

    def _recv_command(self):
        """Receives a command via zmq and updates the command buffer"""
        self.command_buffer[0] = self.command_socket.recv_json()
        ts = time.time()
        if abs(ts - self.command_buffer[0][0]['ts']) > self._time_stamp_tolerance:
            warnings.warn('CommandReceiverThread desynchronized.', RuntimeWarning)

    def run(self):
        while not self.exit_event.is_set():
            # use try-except block without timeout defined above to
            # allow thread to exit even if no event is ever received
            try:
                self._recv_command()
            except:
                pass
        print '[INFO] CommandReceiverThread shutting down.'

    def done(self):
        return self.done_buffer[0]



# -*- coding: utf-8 -*-
import re
from setuptools import setup

# determine version from __init__.py without importing it
with open('./gymz/__init__.py', 'r') as f:
    for l in f:
        if '__version__' in l:
            try:
                version = re.compile('[0-9]+.[0-9]+.[0-9]+').search(l).group()
            except AttributeError:
                raise ValueError('Could not determine package version.')
            else:
                break

setup(
    name='gymz',
    version=version,
    author='Jakob Jordan, Philipp Weidel',
    author_email='j.jordan@fz-juelich.de',
    description=('A light-weight ZeroMQ wrapper for the OpenAI Gym.'),
    license='MIT',
    keywords='openai-gym reinforcement-learning zeromq zmq',
    url='https://github.com/INM-6/python-gymz',
    packages=['gymz'],
    package_data={
        'gymz': ['DefaultConfig.json']
    },
    scripts=['gymz-controller'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering',
    ],
    install_requires=[
        'docopt',
        'gym>=0.8.1',
        'numpy',
        'pyzmq',
    ]
)



# -*- coding: utf-8 -*-

import numpy as np
import os
import time
import warnings

import gym  # openai gym
import gym.spaces
import gym.wrappers

from .wrapper_base import WrapperBase
from . import messages


class GymWrapper(WrapperBase):
    """Wrapper for the OpenAI Gym toolkit"""

    def __init__(self, config):
        WrapperBase.__init__(self)

        self._initial_reward = config['Env']['initial_reward']
        try:
            self._final_reward = config['Env']['final_reward']
        except KeyError:
            self._final_reward = None
        try:
            self._final_reward_null = config['Env']['final_reward_null']
        except KeyError:
            self._final_reward_null = None

        self._min_reward = config['Env']['min_reward']
        self._max_reward = config['Env']['max_reward']
        self._inter_trial_observation = config['Env']['inter_trial_observation']
        self._render = config['Env']['render']

        self._episode = 0
        self._episode_step = 0
        self._episode_reward = []
        self._episode_observation = []
        self._episode_time_start = time.time()

        self._monitor = config['Env']['monitor']
        if self._monitor:
            self._monitor_dir = os.path.join(config['All']['prefix'], config['Env']['monitor_dir'])

    def seed(self, seed):
        self._env.seed(seed)

    def load_env(self, env, *args, **kwargs):
        if 'monitor_args' in kwargs:  # handle monitor args separately
            monitor_args = kwargs['monitor_args']
            del kwargs['monitor_args']

            if not self._monitor and len(monitor_args) > 0:
                warnings.warn('Monitoring not enabled but passing monitor arguments.', RuntimeWarning)

        self._env = gym.make(env, *args, **kwargs)
        self._check_parameters()

        if self._monitor:
            self._env = gym.wrappers.Monitor(self._env, self._monitor_dir, **monitor_args)

    def _check_parameters(self):
        if self._min_reward is not None and np.shape(self._min_reward) != ():
            raise ValueError('min_reward needs to be one dimensional. Please adjust your config.')
        if self._max_reward is not None and np.shape(self._max_reward) != ():
            raise ValueError('max_reward needs to be one dimensional. Please adjust your config.')
        if self._initial_reward is not None and np.shape(self._initial_reward) != ():
            raise ValueError('initial_reward needs to be one dimensional. Please adjust your config.')
        if self._final_reward is not None and np.shape(self._final_reward) != ():
            raise ValueError('final_reward needs to be one dimensional. Please adjust your config.')
        if self._final_reward_null is not None and np.shape(self._final_reward_null) != ():
            raise ValueError('final_reward_null needs to be one dimensional. Please adjust your config.')
        if np.shape(self._inter_trial_observation) != np.shape(self._env.observation_space.sample()):
            raise ValueError('inter_trial_observation is {} dimensional while the obervation space is {} dimensional. These need to be equal. Please adjust your config.'.format(len(np.shape(self._inter_trial_observation)), len(np.shape(self._env.observation_space.sample()))))

    def reset(self):
        self._output = self._env.reset()  # reset returns initial state
        self._done_buffer[0] = False

        # initial reward is always assumed to be zero (by Gym), we
        # allow to overwrite it with a custom value to avoid potential
        # jumps in value from first to second state
        self._reward = self._initial_reward

        self._episode += 1
        self._episode_step = 0
        self._episode_reward = []
        self._episode_observation = []
        self._episode_time_start = time.time()

    def execute_action(self):
        # Gym expects actions in different format depending on type of
        # action space
        if isinstance(self._env.action_space, gym.spaces.Discrete):
            action = self._command_buffer[0][0]['value']
        else:
            action = [self._command_buffer[0][0]['value']]

        self._output, self._reward, self._done_buffer[0], _ = self._env.step(action)

        # record Gym output depending on type of observation space
        if isinstance(self._env.observation_space, gym.spaces.Discrete):
            self._episode_observation.append(self._output)
        elif isinstance(self._env.observation_space, gym.spaces.Box):
            self._episode_observation.append(list(self._output))
        else:
            raise NotImplementedError('Observation space {obs} not supported.'.format(obs=self._env.observation_space))

        self._episode_step += 1

        # in case user provides a reward value, overwrite the one retrieved from Gym
        if self._done_buffer[0] is True:
            if self._final_reward is not None:
                self._reward = self._final_reward
            elif abs(self._reward) < 1e-10:
                self._reward = self._final_reward_null

        self._episode_reward.append(self._reward)

        if self._render:
            self._env.render()

    def update_output_buffer(self):
        assert(self._output_buffer is not None)

        # handle Gym output depending on type of observation space
        if isinstance(self._env.observation_space, gym.spaces.Discrete):
            self._output_buffer[0] = messages.to_message(0, self._env.observation_space.n - 1, self._output)
        elif isinstance(self._env.observation_space, gym.spaces.Box):
            self._output_buffer[0] = messages.to_message(self._env.observation_space.low, self._env.observation_space.high, self._output)
        else:
            raise NotImplementedError('Observation space {obs} not supported.'.format(obs=self._env.observation_space))

    def get_command_buffer(self):
        if self._command_buffer is None:
            # set up buffer depending on type of action space
            if isinstance(self._env.action_space, gym.spaces.Discrete):
                self._command_buffer = [messages.to_message(0, self._env.action_space.n - 1, 0)]
            elif isinstance(self._env.action_space, gym.spaces.Box) and len(self._env.action_space.shape) == 1:
                self._command_buffer = [messages.to_message(self._env.action_space.low, self._env.action_space.high, 0.)]
            else:
                raise NotImplementedError('Action space {acts} not supported.'.format(acts=self._env.action_space))
        return self._command_buffer

    def get_output_buffer(self):
        if self._output_buffer is None:
            self._output_buffer = [[]]
            self._output = self._env.reset()  # reset returns initial observation
            self.update_output_buffer()
        return self._output_buffer

    def clear_output_buffer(self):
        assert(self._output_buffer is not None)

        if isinstance(self._env.observation_space, gym.spaces.Discrete):
            self._output_buffer[0] = messages.to_message(0, self._env.observation_space.n - 1, self._inter_trial_observation)
        elif isinstance(self._env.observation_space, gym.spaces.Box) and len(self._env.observation_space.shape) == 1:
            if np.shape(self._env.observation_space.low) != np.shape(self._inter_trial_observation):
                raise ValueError('Dimensions of inter_trial_observation do not match environment.')
            self._output_buffer[0] = messages.to_message(self._env.observation_space.low, self._env.observation_space.high, self._inter_trial_observation)
        else:
            raise NotImplementedError('Observation space {obs} not supported.'.format(obs=self._env.observation_space))

    def report(self):
        return {
            self._episode: {
                'reward': self._episode_reward,
                'obervation': self._episode_observation,
                'duration': time.time() - self._episode_time_start
            }
        }



# -*- coding: utf-8 -*-

import json
import os
import threading
import time
import warnings

from . import misc


class EnvRunnerThread(threading.Thread):
    """Runs the environment."""

    def __init__(self, thread_id, thread_name, emu, command_buffer, output_buffer, reward_buffer, config, exit_event):
        threading.Thread.__init__(self)
        self.thread_id = thread_id
        self.name = thread_name
        self.emu = emu
        self.command_buffer = command_buffer
        self.output_buffer = output_buffer
        self.reward_buffer = reward_buffer
        self.exit_event = exit_event
        self._update_interval = config['EnvRunner']['update_interval']  # (s), update interval of ZMQ sockets
        self._inter_trial_duration = config['EnvRunner']['inter_trial_duration']  # (s), duration of no input/no reward state after reset of environment
        self._write_report = config['All']['write_report']
        if self._write_report:
            self._report_file = os.path.join(config['All']['prefix'], config['All']['report_file'])
            self._flush_report_interval = config['All']['flush_report_interval']

            if os.path.isfile(self._report_file):  # file already exists
                if config['All']['overwrite_files']:
                    warnings.warn('Report file already exists. Truncating.', RuntimeWarning)
                    with open(self._report_file, 'w') as f:  # clear file
                        json.dump({}, f)
                else:
                    raise IOError('Report file already exists. Exiting.')
        else:
            self._report_file = None

    def _report(self):
        if not self._write_report:
            return

        try:
            with open(self._report_file, 'r') as f:
                report = json.load(f)
        except IOError:
            report = {}

        report.update(self.emu.report())
        with open(self._report_file, 'w') as f:
            json.dump(report, f)

    def run(self):
        # Reset environment
        self.emu.reset()

        # Run environment
        step = 0
        while not self.exit_event.is_set():
            t_start = time.time()

            if self.emu.done():
                print '[info] EnvRunnerThread: reset'
                t_start_done = time.time()

                if self._write_report and self._flush_report_interval is None:
                    self._report()

                # reset environment
                self.emu.reset()

                # clear all input for a certain duration
                self.emu.clear_output_buffer()
                self.emu.clear_reward_buffer()
                misc.sleep_remaining(t_start_done, self._inter_trial_duration,
                                     'EnvRunnerThread: inter trial sleep time negative')
                t_start = time.time()

                # update buffers to reflect initial state of env
                self.emu.update_output_buffer()
                self.emu.update_reward_buffer()
            else:
                self.emu.execute_action()
                self.emu.update_output_buffer()
                self.emu.update_reward_buffer()

            step += 1

            if self._write_report and self._flush_report_interval is not None and step % self._flush_report_interval == 0:
                self._report()

            misc.sleep_remaining(t_start, self._update_interval,
                                 'EnvRunnerThread: sleep time negative')
        print '[INFO] EnvRunnerThread shutting down.'
